#ifndef __VECTOR_H__
#define __VECTOR_H__

#include "mathlibx.h"

//////////////////////////////////////////////////
// Operations Supported
//
// Vector u, v, w;   // Vectors
// Point p;          // Points
// FLOAT f;          // Scalars
// BOOL b;           // Booleans
//
// f = v[0];       // LHS Index Operation
// v[0] = f;       // LHS Index Operation
// f = v(0);       // RHS Index Operation
// v(0) = f;       // ILLEGAL -- RHS (right hand side) only!!!
//
// u = v;
// u = p;
// u += v;
// u -= v;
// u *= f;         // Scalar Multiplication
// u /= f;         // Scalar Division
// u = -v;
// u = v + w;
// u = v - w;
// u = v * w;      // Cross Product
// u = v.Cross(w); // Cross Product
// f = v.Dot(w);   // Dot Product
// f = v.Dot(p);   // Dot Product for on Plane Calculations
// u = v * f;      // Scalar Multiplication
// u = f * v;      // Scalar Multiplication
// u = v / f;      // Scalar Division
// u = f / v;      // Scalar Inversion
//
// Logic
// b = (u == v); // Quasi Equal allows for floating point inaccuracy
// b = (u != v); // Quasi Not Equal allows for floating point inaccuracy

//////////////////////////////////////////////////////////////////////
// Vector Class
//

class Point;

class Vector
{
public:
  Vector();
  Vector(FLOAT x, FLOAT y, FLOAT z, FLOAT w = 0.0f);
  virtual ~Vector();

public:
  inline VOID Set(FLOAT x, FLOAT y, FLOAT z, FLOAT w = 0.0f);
  inline VOID Set(const VECTOR4D v);
  inline VOID Get(VECTOR4D v) const;
  inline const FLOAT *Get() const;
  virtual inline BOOL GetDynamic();

public:
  virtual VOID Update();

public:
  FLOAT Magnitude() const;
  VOID Null();
  BOOL Normalize();
  VOID CompletePlaneEquation(const Point &pt);

public:
  inline FLOAT &operator[](const UINT_32 uiIndex);
  inline FLOAT operator()(const UINT_32 uiIndex) const;

  Vector &operator=(const Vector &v);
  Vector &operator=(const Point &pt);
  Vector &operator+=(const Vector &v);
  Vector &operator-=(const Vector &v);
  Vector &operator*=(const FLOAT f);
  Vector &operator/=(const FLOAT f);

  BOOL operator==(const Vector &v) const;
  BOOL operator!=(const Vector &v) const;

  inline Vector Cross(const Vector &v) const;
  FLOAT Dot(const Vector &v) const;

public:
  VOID Display() const;

protected:
  VECTOR4D m_v;
};

//////////////////////////////////////////////////////////////////////
// External Operators
//

inline Vector operator-(const Vector &v);
inline Vector operator+(const Vector &v, const Vector &w);
inline Vector operator-(const Vector &v, const Vector &w);
inline Vector operator*(const Vector &v, const Vector &w);
inline Vector operator*(const Vector &v, const FLOAT f);
inline Vector operator*(const FLOAT f, const Vector &v);
inline Vector operator/(const Vector &v, const FLOAT f);
inline Vector operator/(const FLOAT f, const Vector &v);

inline Vector ComponentMultiply(const Vector &v, const Vector &w);
DOUBLE AngleBetween(const Vector &v1, const Vector &v2);

//////////////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID Vector::Set(FLOAT x, FLOAT y, FLOAT z, FLOAT w/* = 0.0f*/)
{
  m_v[0] = x;
  m_v[1] = y;
  m_v[2] = z;
  m_v[3] = w;
}

inline VOID Vector::Set(const VECTOR4D v)
{
  ASSERT( v != NULL );

  memcpy(m_v, v, sizeof(VECTOR4D));
}

inline VOID Vector::Get(VECTOR4D v) const
{
  ASSERT( v != NULL );

  memcpy(v, m_v, sizeof(VECTOR4D));
}

inline const FLOAT *Vector::Get() const
{
  return m_v;
}

inline BOOL Vector::GetDynamic()
{
  return FALSE;
}

inline FLOAT &Vector::operator[](const UINT_32 uiIndex)
{
  return m_v[uiIndex];
}

inline FLOAT Vector::operator()(const UINT_32 uiIndex) const
{
  return m_v[uiIndex];
}

inline Vector Vector::Cross(const Vector &v) const
{
  return (*this) * v;
}

//////////////////////////////////////////////////////////////////////
// External Operator Implementations
//

inline Vector operator-(const Vector &v)
{
  return Vector(-v(0), -v(1), -v(2), -v(3));
}

inline Vector operator+(const Vector &v, const Vector &w)
{
  return Vector(v(0) + w(0), v(1) + w(1), v(2) + w(2), v(3) + w(3));
}

inline Vector operator-(const Vector &v, const Vector &w)
{
  return Vector(v(0) - w(0), v(1) - w(1), v(2) - w(2), v(3) - w(3));
}

inline Vector operator*(const Vector &v, const Vector &w)
{
  return Vector(v(1)*w(2) - v(2)*w(1),
		v(2)*w(0) - v(0)*w(2), 
		v(0)*w(1) - v(1)*w(0),  
		0.0);
}

inline Vector operator*(const Vector &v, const FLOAT f)
{
  return Vector(v(0)*f, v(1)*f, v(2)*f, v(3)*f);  
}

inline Vector operator*(const FLOAT f, const Vector &v)
{
  return Vector(v(0)*f, v(1)*f, v(2)*f, v(3)*f);  
}

inline Vector operator/(const Vector &v, const FLOAT f)
{
  ASSERT( f != 0.0 );

  return Vector(v(0)/f, v(1)/f, v(2)/f, v(3)/f);  
}

inline Vector operator/(const FLOAT f, const Vector &v)
{
  return Vector((v(0) == 0.0) ? ((FLOAT)0.0) : f/v(0),
		(v(1) == 0.0) ? ((FLOAT)0.0) : f/v(1),
		(v(2) == 0.0) ? ((FLOAT)0.0) : f/v(2),
		(v(3) == 0.0) ? ((FLOAT)0.0) : f/v(3));
}

inline Vector ComponentMultiply(const Vector &v, const Vector &w)
{
  return Vector(v(0)*w(0), v(1)*w(1), v(2)*w(2), v(3)*w(3));  
}

//////////////////////////////////////////////////////////////////////
// DynVector Class
//

class DynVector : public Vector
{
public:
  DynVector();
  DynVector(FLOAT x, FLOAT y, FLOAT z, FLOAT w = 0.0f);
  virtual ~DynVector();

public:
  inline VOID SetAll(DynVector *pdynv);
  inline VOID SetAll(DynFloat *pdynfX, DynFloat *pdynfY = NULL, 
		     DynFloat *pdynfZ = NULL, DynFloat *pdynfW = NULL);
  inline VOID SetDynExprs(CHAR *pcX, CHAR *pcY = NULL, 
                          CHAR *pcZ = NULL, CHAR *pcW = NULL);
  virtual inline BOOL GetDynamic();

public:
  virtual VOID Update();

public:
  DynVector &operator=(const Vector &v);

protected:
  CHAR *m_ppcDynExpr[4];
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID DynVector::SetAll(DynVector *pdynv)
{
  ASSERT( pdynv != NULL );

  memcpy(this, pdynv, sizeof(DynVector));
}

inline VOID DynVector::SetAll(DynFloat *pdynfX, 
			      DynFloat *pdynfY/* = NULL*/,
			      DynFloat *pdynfZ/* = NULL*/, 
			      DynFloat *pdynfW/* = NULL*/)
{
  Null();
  if ( pdynfX != NULL )
    {
      m_v[0] = pdynfX->GetValue();
      m_ppcDynExpr[0] = pdynfX->GetDynExpr();
    }
  if ( pdynfY != NULL )
    {
      m_v[1] = pdynfY->GetValue();
      m_ppcDynExpr[1] = pdynfY->GetDynExpr();
    }
  if ( pdynfZ != NULL )
    {
      m_v[2] = pdynfZ->GetValue();
      m_ppcDynExpr[2] = pdynfZ->GetDynExpr();
    }
  if ( pdynfW != NULL )
    {
      m_v[3] = pdynfW->GetValue();
      m_ppcDynExpr[3] = pdynfW->GetDynExpr();
    }
}

inline VOID DynVector::SetDynExprs(CHAR *pcX, 
				   CHAR *pcY/* = NULL*/, 
                                   CHAR *pcZ/* = NULL*/, 
				   CHAR *pcW/* = NULL*/)
{
  m_ppcDynExpr[0] = pcX;
  m_ppcDynExpr[1] = pcY;
  m_ppcDynExpr[2] = pcZ;
  m_ppcDynExpr[3] = pcW;
}

inline BOOL DynVector::GetDynamic()
{
  return ( m_ppcDynExpr[0] != NULL ||
	   m_ppcDynExpr[1] != NULL ||
	   m_ppcDynExpr[2] != NULL ||
	   m_ppcDynExpr[3] != NULL );
}

#endif // __VECTOR_H__

